from typing import TYPE_CHECKING, Any, Container, Dict, List, Optional, Sized, Tuple, Type, TypeVar, Union, overload

from ._base import DirtyEquals
from ._utils import Omit, plain_repr

if TYPE_CHECKING:
    from typing import TypeAlias

__all__ = 'HasLen', 'Contains', 'IsListOrTuple', 'IsList', 'IsTuple'
T = TypeVar('T', List[Any], Tuple[Any, ...])
LengthType: 'TypeAlias' = 'Union[None, int, Tuple[int, Union[int, Any]]]'


class HasLen(DirtyEquals[Sized]):
    """
    Check that some has a given length, or length in a given range.
    """

    @overload
    def __init__(self, length: int):
        ...

    @overload
    def __init__(self, min_length: int, max_length: Union[int, Any]):
        ...

    def __init__(self, min_length: int, max_length: Union[None, int, Any] = None):  # type: ignore[misc]
        """
        Args:
            min_length: Expected length if `max_length` is not given, else minimum length.
            max_length: Expected maximum length, use an ellipsis `...` to indicate that there's no maximum.

        ```py title="HasLen"
        from dirty_equals import HasLen

        assert [1, 2, 3] == HasLen(3) #(1)!
        assert '123' == HasLen(3, ...) #(2)!
        assert (1, 2, 3) == HasLen(3, 5) #(3)!
        assert (1, 2, 3) == HasLen(0, ...) #(4)!
        ```

        1. Length must be 3.
        2. Length must be 3 or higher.
        3. Length must be between 3 and 5 inclusive.
        4. Length is required but can take any value.
        """
        if max_length is None:
            self.length: 'LengthType' = min_length
            super().__init__(self.length)
        else:
            self.length = (min_length, max_length)
            super().__init__(*_length_repr(self.length))

    def equals(self, other: Any) -> bool:
        return _length_correct(self.length, other)


class Contains(DirtyEquals[Container[Any]]):
    """
    Check that an object contains one or more values.
    """

    def __init__(self, contained_value: Any, *more_contained_values: Any):
        """
        Args:
            contained_value: value that must be contained in the compared object.
            *more_contained_values: more values that must be contained in the compared object.

        ```py title="Contains"
        from dirty_equals import Contains

        assert [1, 2, 3] == Contains(1)
        assert [1, 2, 3] == Contains(1, 2)
        assert (1, 2, 3) == Contains(1)
        assert 'abc' == Contains('b')
        assert {'a': 1, 'b': 2} == Contains('a')
        assert [1, 2, 3] != Contains(10)
        ```
        """
        self.contained_values: Tuple[Any, ...] = (contained_value,) + more_contained_values
        super().__init__(*self.contained_values)

    def equals(self, other: Any) -> bool:
        return all(v in other for v in self.contained_values)


class IsListOrTuple(DirtyEquals[T]):
    """
    Check that some object is a list or tuple and optionally its values match some constraints.
    """

    allowed_type: Union[Type[T], Tuple[Type[List[Any]], Type[Tuple[Any, ...]]]] = (list, tuple)

    @overload
    def __init__(self, *items: Any, check_order: bool = True, length: 'LengthType' = None):
        ...

    @overload
    def __init__(self, positions: Dict[int, Any], length: 'LengthType' = None):
        ...

    def __init__(
        self,
        *items: Any,
        positions: Optional[Dict[int, Any]] = None,
        check_order: bool = True,
        length: 'LengthType' = None,
    ):
        """
        `IsListOrTuple` and its subclasses can be initialised in two ways:

        Args:
            *items: Positional members of an object to check. These must start from the zeroth position, but
                (depending on the value of `length`) may not include all values of the list/tuple being checked.
            check_order: Whether to enforce the order of the items.
            length (Union[int, Tuple[int, Union[int, Any]]]): length constraints, int or tuple matching the arguments
                of [`HasLen`][dirty_equals.HasLen].

        or,

        Args:
            positions (Dict[int, Any]): Instead of `*items`, a dictionary of positions and
                values to check and be provided.
            length (Union[int, Tuple[int, Union[int, Any]]]): length constraints, int or tuple matching the arguments
                of [`HasLen`][dirty_equals.HasLen].

        ```py title="IsListOrTuple"
        from dirty_equals import IsListOrTuple, AnyThing

        assert [1, 2, 3] == IsListOrTuple(1, 2, 3)
        assert (1, 3, 2) == IsListOrTuple(1, 2, 3, check_order=False)
        assert [{'a': 1}, {'a': 2}] == (
            IsListOrTuple({'a': 2}, {'a': 1}, check_order=False) #(1)!
        )
        assert [1, 2, 3, 3] != IsListOrTuple(1, 2, 3, check_order=False) #(2)!

        assert [1, 2, 3, 4, 5] == IsListOrTuple(1, 2, 3, length=...) #(3)!
        assert [1, 2, 3, 4, 5] != IsListOrTuple(1, 2, 3, length=(8, 10)) #(4)!

        assert ['a', 'b', 'c', 'd'] == (
            IsListOrTuple(positions={2: 'c', 3: 'd'}) #(5)!
        )
        assert ['a', 'b', 'c', 'd'] == (
            IsListOrTuple(positions={2: 'c', 3: 'd'}, length=4) #(6)!
        )

        assert [1, 2, 3, 4] == IsListOrTuple(3, check_order=False, length=(0, ...)) #(7)!

        assert [1, 2, 3] == IsListOrTuple(AnyThing, AnyThing, 3) #(8)!
        ```

        1. Unlike using sets for comparison, we can do order-insensitive comparisons on objects that are not hashable.
        2. And we won't get caught out by duplicate values
        3. Here we're just checking the first 3 items, the compared list or tuple can be of any length
        4. Compared list is not long enough
        5. Compare using `positions`, here no length if enforced
        6. Compare using `positions` but with a length constraint
        7. Here we're just confirming that the value `3` is in the list
        8. If you don't care about the first few values of a list or tuple,
            you can use [`AnyThing`][dirty_equals.AnyThing] in your arguments.
        """
        if positions is not None:
            self.positions: Optional[Dict[int, Any]] = positions
            if items:
                raise TypeError(f'{self.__class__.__name__} requires either args or positions, not both')
            if not check_order:
                raise TypeError('check_order=False is not compatible with positions')
        else:
            self.positions = None
            self.items = items
        self.check_order = check_order

        self.length: Any = length
        if self.length is not None and not isinstance(self.length, int):
            if self.length == Ellipsis:
                self.length = 0, ...
            else:
                self.length = tuple(self.length)

        super().__init__(
            *items,
            positions=Omit if positions is None else positions,
            length=_length_repr(self.length),
            check_order=self.check_order and Omit,
        )

    def equals(self, other: Any) -> bool:  # noqa: C901
        if not isinstance(other, self.allowed_type):
            return False

        if not _length_correct(self.length, other):
            return False

        if self.check_order:
            if self.positions is None:
                if self.length is None:
                    return list(self.items) == list(other)
                else:
                    return list(self.items) == list(other[: len(self.items)])
            else:
                return all(v == other[k] for k, v in self.positions.items())
        else:
            # order insensitive comparison
            # if we haven't checked length yet, check it now
            if self.length is None and len(other) != len(self.items):
                return False

            other_copy = list(other)
            for item in self.items:
                try:
                    other_copy.remove(item)
                except ValueError:
                    return False
            return True


class IsList(IsListOrTuple[List[Any]]):
    """
    All the same functionality as [`IsListOrTuple`][dirty_equals.IsListOrTuple], but the compared value must be a list.

    ```py title="IsList"
    from dirty_equals import IsList

    assert [1, 2, 3] == IsList(1, 2, 3)
    assert [1, 2, 3] == IsList(positions={2: 3})
    assert [1, 2, 3] == IsList(1, 2, 3, check_order=False)
    assert [1, 2, 3, 4] == IsList(1, 2, 3, length=4)
    assert [1, 2, 3, 4] == IsList(1, 2, 3, length=(4, 5))
    assert [1, 2, 3, 4] == IsList(1, 2, 3, length=...)

    assert (1, 2, 3) != IsList(1, 2, 3)
    ```
    """

    allowed_type = list


class IsTuple(IsListOrTuple[Tuple[Any, ...]]):
    """
    All the same functionality as [`IsListOrTuple`][dirty_equals.IsListOrTuple], but the compared value must be a tuple.

    ```py title="IsTuple"
    from dirty_equals import IsTuple

    assert (1, 2, 3) == IsTuple(1, 2, 3)
    assert (1, 2, 3) == IsTuple(positions={2: 3})
    assert (1, 2, 3) == IsTuple(1, 2, 3, check_order=False)
    assert (1, 2, 3, 4) == IsTuple(1, 2, 3, length=4)
    assert (1, 2, 3, 4) == IsTuple(1, 2, 3, length=(4, 5))
    assert (1, 2, 3, 4) == IsTuple(1, 2, 3, length=...)

    assert [1, 2, 3] != IsTuple(1, 2, 3)
    ```
    """

    allowed_type = tuple


def _length_repr(length: 'LengthType') -> Any:
    if length is None:
        return Omit
    elif isinstance(length, int):
        return length
    else:
        if len(length) != 2:
            raise TypeError(f'length must be a tuple of length 2, not {len(length)}')
        max_value = length[1] if isinstance(length[1], int) else plain_repr('...')
        return length[0], max_value


def _length_correct(length: 'LengthType', other: 'Sized') -> bool:
    if isinstance(length, int):
        if len(other) != length:
            return False
    elif isinstance(length, tuple):
        other_len = len(other)
        min_length, max_length = length
        if other_len < min_length:
            return False
        if isinstance(max_length, int) and other_len > max_length:
            return False
    return True



from ._base import AnyThing, DirtyEquals, IsOneOf
from ._boolean import IsFalseLike, IsTrueLike
from ._datetime import IsDate, IsDatetime, IsNow, IsToday
from ._dict import IsDict, IsIgnoreDict, IsPartialDict, IsStrictDict
from ._inspection import HasAttributes, HasName, HasRepr, IsInstance
from ._numeric import (
    IsApprox,
    IsFloat,
    IsInt,
    IsNegative,
    IsNegativeFloat,
    IsNegativeInt,
    IsNonNegative,
    IsNonPositive,
    IsNumber,
    IsNumeric,
    IsPositive,
    IsPositiveFloat,
    IsPositiveInt,
)
from ._other import FunctionCheck, IsJson, IsUUID
from ._sequence import Contains, HasLen, IsList, IsListOrTuple, IsTuple
from ._strings import IsAnyStr, IsBytes, IsStr

__all__ = (
    # base
    'DirtyEquals',
    'AnyThing',
    'IsOneOf',
    # boolean
    'IsTrueLike',
    'IsFalseLike',
    # datetime
    'IsDatetime',
    'IsNow',
    'IsDate',
    'IsToday',
    # dict
    'IsDict',
    'IsPartialDict',
    'IsIgnoreDict',
    'IsStrictDict',
    # sequence
    'Contains',
    'HasLen',
    'IsList',
    'IsTuple',
    'IsListOrTuple',
    # numeric
    'IsNumeric',
    'IsApprox',
    'IsNumber',
    'IsPositive',
    'IsNegative',
    'IsNonPositive',
    'IsNonNegative',
    'IsInt',
    'IsPositiveInt',
    'IsNegativeInt',
    'IsFloat',
    'IsPositiveFloat',
    'IsNegativeFloat',
    # inspection
    'HasAttributes',
    'HasName',
    'HasRepr',
    'IsInstance',
    # other
    'FunctionCheck',
    'IsJson',
    'IsUUID',
    # strings
    'IsStr',
    'IsBytes',
    'IsAnyStr',
    # version
    '__version__',
)

__version__ = '0.4'



import json
from typing import Any, Callable, TypeVar, overload
from uuid import UUID

from ._base import DirtyEquals
from ._utils import plain_repr

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore[misc]


class IsUUID(DirtyEquals[UUID]):
    """
    A class that checks if a value is a valid UUID, optionally checking UUID version.
    """

    def __init__(self, version: Literal[None, 1, 2, 3, 4, 5] = None):
        """
        Args:
            version: The version of the UUID to check, if omitted, all versions are accepted.

        ```py title="IsUUID"
        import uuid
        from dirty_equals import IsUUID

        assert 'edf9f29e-45c7-431c-99db-28ea44df9785' == IsUUID
        assert 'edf9f29e-45c7-431c-99db-28ea44df9785' == IsUUID(4)
        assert 'edf9f29e45c7431c99db28ea44df9785' == IsUUID(4)
        assert 'edf9f29e-45c7-431c-99db-28ea44df9785' != IsUUID(5)
        assert uuid.uuid4() == IsUUID(4)
        ```
        """
        self.version = version
        super().__init__(version or plain_repr('*'))

    def equals(self, other: Any) -> bool:
        if isinstance(other, UUID):
            uuid = other
        elif isinstance(other, str):
            uuid = UUID(other)
            if self.version is not None and uuid.version != self.version:
                return False
        else:
            return False

        if self.version:
            return uuid.version == self.version
        else:
            return True


AnyJson = object
JsonType = TypeVar('JsonType', AnyJson, Any)


class IsJson(DirtyEquals[JsonType]):
    """
    A class that checks if a value is a JSON object, and check the contents of the JSON.
    """

    @overload
    def __init__(self, expected_value: JsonType = AnyJson):
        ...

    @overload
    def __init__(self, **expected_kwargs: Any):
        ...

    def __init__(self, expected_value: JsonType = AnyJson, **expected_kwargs: Any):
        """
        Args:
            expected_value: Value to compare the JSON to, if omitted, any JSON is accepted.
            **expected_kwargs: Keyword arguments forming a dict to compare the JSON to,
                `expected_value` and `expected_kwargs` may not be combined.

        As with any `dirty_equals` type, types can be nested to provide more complex checks.

        !!! note
            Like [`IsInstance`][dirty_equals.IsInstance], `IsJson` can be parameterized or initialised with a value -
            `IsJson[xyz]` is exactly equivalent to `IsJson(xyz)`.

            This allows usage to be analogous to type hints.


        ```py title="IsJson"
        from dirty_equals import IsJson, IsStrictDict, IsPositiveInt

        assert '{"a": 1, "b": 2}' == IsJson
        assert '{"a": 1, "b": 2}' == IsJson(a=1, b=2)
        assert '{"a": 1}' != IsJson(a=2)
        assert 'invalid json' != IsJson
        assert '{"a": 1}' == IsJson(a=IsPositiveInt)
        assert '"just a quoted string"' == IsJson('just a quoted string')

        assert '{"a": 1, "b": 2}' == IsJson[IsStrictDict(a=1, b=2)]
        assert '{"b": 2, "a": 1}' != IsJson[IsStrictDict(a=1, b=2)]
        ```
        """
        if expected_kwargs:
            if expected_value is not AnyJson:
                raise TypeError('IsJson requires either an argument or kwargs, not both')
            self.expected_value: Any = expected_kwargs
        else:
            self.expected_value = expected_value
        super().__init__(plain_repr('*') if expected_value is AnyJson else expected_value)

    def __class_getitem__(cls, expected_type: JsonType) -> 'IsJson[JsonType]':
        return cls(expected_type)

    def equals(self, other: Any) -> bool:
        if isinstance(other, (str, bytes)):
            v = json.loads(other)
            if self.expected_value is AnyJson:
                return True
            else:
                return v == self.expected_value
        else:
            return False


class FunctionCheck(DirtyEquals[Any]):
    """
    Use a function to check if a value "equals" whatever you want to check
    """

    def __init__(self, func: Callable[[Any], bool]):
        """
        Args:
            func: callable that takes a value and returns a bool.

        ```py title="FunctionCheck"
        from dirty_equals import FunctionCheck

        def is_even(x):
            return x % 2 == 0

        assert 2 == FunctionCheck(is_even)
        assert 3 != FunctionCheck(is_even)
        ```
        """
        self.func = func
        super().__init__(plain_repr(func.__name__))

    def equals(self, other: Any) -> bool:
        return self.func(other)



from typing import Any

from ._base import DirtyEquals
from ._utils import Omit


class IsTrueLike(DirtyEquals[bool]):
    """
    Check if the value is True like. `IsTrueLike` allows comparison to anything and effectively uses just
    `return bool(other)`.

    Example of basic usage:

    ```py title="IsTrueLike"
    from dirty_equals import IsTrueLike

    assert True == IsTrueLike
    assert 1 == IsTrueLike
    assert 'true' == IsTrueLike
    assert 'foobar' == IsTrueLike  # any non-empty string is "True"
    assert '' != IsTrueLike
    assert [1] == IsTrueLike
    assert {} != IsTrueLike
    assert None != IsTrueLike
    ```
    """

    def equals(self, other: Any) -> bool:
        return bool(other)


class IsFalseLike(DirtyEquals[bool]):
    """
    Check if the value is False like. `IsFalseLike` allows comparison to anything and effectively uses
    `return not bool(other)` (with string checks if `allow_strings=True` is set).
    """

    def __init__(self, *, allow_strings: bool = False):
        """
        Args:
            allow_strings: if `True`, allow comparisons to `False` like strings, case-insensitive, allows
                `''`, `'false'` and any string where `float(other) == 0` (e.g. `'0'`).

        Example of basic usage:

        ```py title="IsFalseLike"
        from dirty_equals import IsFalseLike

        assert False == IsFalseLike
        assert 0 == IsFalseLike
        assert 'false' == IsFalseLike(allow_strings=True)
        assert '0' == IsFalseLike(allow_strings=True)
        assert 'foobar' != IsFalseLike(allow_strings=True)
        assert 'false' != IsFalseLike
        assert 'True' != IsFalseLike(allow_strings=True)
        assert [1] != IsFalseLike
        assert {} == IsFalseLike
        assert None == IsFalseLike
        assert '' == IsFalseLike(allow_strings=True)
        assert '' == IsFalseLike
        ```
        """
        self.allow_strings = allow_strings
        super().__init__(allow_strings=allow_strings or Omit)

    def equals(self, other: Any) -> bool:
        if isinstance(other, str) and self.allow_strings:
            return self.make_string_check(other)
        return not bool(other)

    @staticmethod
    def make_string_check(other: str) -> bool:
        if other.lower() in {'false', ''}:
            return True

        try:
            return float(other) == 0
        except ValueError:
            return False
